"""Model for the Ecoforest stove status."""
from __future__ import annotations

from dataclasses import dataclass
from enum import Enum

from pyecoforest.const import MODEL_NAME, SUPPORTED_MODELS
from pyecoforest.exceptions import EcoforestError


class OperationMode(Enum):
    """Model that represents the device operation modes."""

    POWER = "power"
    TEMPERATURE = "temperature"
    EMERGENCY = "emergency"

    @classmethod
    def build(cls, mode: str) -> OperationMode:
        """Parse the operation mode code to an OperationMode object."""
        modes = {
            "0": OperationMode.POWER,
            "1": OperationMode.TEMPERATURE,
            "2": OperationMode.EMERGENCY,
        }

        if mode in modes:
            return modes[mode]

        raise EcoforestError(f"The operation mode {mode} is not a valid operation!")


class State(Enum):
    """Model that represents the state of the device."""

    OFF = "off"
    STARTING = "starting"
    PRE_HEATING = "pre_heating"
    ON = "on"
    SHUTTING_DOWN = "shutting_down"
    STAND_BY = "stand_by"
    ALARM = "alarm"

    @classmethod
    def build(cls, state: str) -> State:
        """Parse the state code to a State object."""
        states = {
            "OFF": [0],
            "STARTING": [1, 2, 3, 4, 10],
            "PRE_HEATING": [5, 6],
            "ON": [7],
            "SHUTTING_DOWN": [8, 11, -3],
            "STAND_BY": [-20],
            "ALARM": [-4],
        }

        for k, v in states.items():
            if int(state) in v:
                return State[k]

        raise EcoforestError(f"The state {state} is not a valid state!")


class Alarm(Enum):
    """Model that represents the alarms of the device."""

    AIR_DEPRESSION = "air_depression"
    PELLETS = "pellets"
    CPU_OVERHEATING = "cpu_overheating"
    UNKNOWN = "unknown"

    @classmethod
    def build(cls, alarm: str) -> Alarm | None:
        """Parse the alarm code to an Alarm object."""
        alarms = {
            "A001": Alarm.AIR_DEPRESSION,
            "A002": Alarm.AIR_DEPRESSION,
            "A012": Alarm.CPU_OVERHEATING,
            "A099": Alarm.PELLETS,
            "N": None,
        }

        if alarm in alarms:
            return alarms[alarm]

        return Alarm.UNKNOWN


@dataclass
class Device:
    """Model for the Ecoforest stove."""

    # model information
    is_supported: bool
    firmware: str
    model: str
    model_name: str
    serial_number: str

    # confirguation information
    operation_mode: OperationMode

    # status information
    on: bool
    state: State
    power: int
    temperature: float
    alarm: Alarm | None = None
    alarm_code: str | None = None

    # sensors
    environment_temperature: float | None = None
    cpu_temperature: float | None = None
    gas_temperature: float | None = None
    ntc_temperature: float | None = None
    depression: int | None = None
    working_hours: int | None = None
    working_state: int | None = None
    working_level: int | None = None
    ignitions: int | None = None
    live_pulse: float | None = None
    pulse_offset: float | None = None
    extractor: float | None = None
    convecto_air_flow: float | None = None

    @classmethod
    def build(cls, data: dict[str, dict[str, str]]) -> Device:
        """Parse request data and return as Device."""
        status = data["status"]
        stats = data["stats"]
        alarms = data["alarms"]
        model = stats["Me"]
        return Device(
            is_supported=model in SUPPORTED_MODELS,
            model=model,
            model_name=MODEL_NAME,
            firmware=stats["Vs"],
            serial_number=stats["Ns"],
            operation_mode=OperationMode.build(status["modo_operacion"]),
            on=status["on_off"] == "1",
            state=State.build(status["estado"]),
            power=int(status["consigna_potencia"]),
            temperature=float(status["consigna_temperatura"]),
            alarm=Alarm.build(alarms["get_alarmas"]),
            alarm_code=alarms["get_alarmas"] if alarms["get_alarmas"] != "N" else None,
            environment_temperature=float(status["temperatura"]),
            cpu_temperature=float(stats["Tp"]),
            gas_temperature=float(stats["Th"]),
            ntc_temperature=float(stats["Tn"]),
            depression=int(stats["Da"]),
            working_hours=int(stats["Nh"]),
            ignitions=int(stats["Ne"]),
            live_pulse=float(stats["Pn"]),
            pulse_offset=float(stats["Pf"]),
            working_state=int(stats["Es"]),
            extractor=float(stats["Ex"]),
            working_level=int(stats["Ni"]),
            convecto_air_flow=float(stats["Co"]),
        )



from pathlib import Path

import httpx
import pytest
import respx

from pyecoforest.api import EcoforestApi
from pyecoforest.const import (
    API_ALARMS_OP,
    API_SET_POWER_OP,
    API_SET_STATE_OP,
    API_SET_TEMP_OP,
    API_STATS_OP,
    API_STATUS_OP,
    URL_CGI,
)
from pyecoforest.exceptions import (
    EcoforestAuthenticationRequired,
    EcoforestConnectionError,
)
from pyecoforest.models.device import Device, OperationMode, State


def _fixtures_dir() -> Path:
    return Path(__file__).parent / "fixtures"


def _load_fixture(name: str) -> str:
    with open(_fixtures_dir() / name) as read_in:
        return read_in.read()


def _mutate_fixture(name: str, pairs: list[tuple[str, str]]) -> str:
    fixture = _load_fixture(name)
    for k, v in pairs:
        fixture = fixture.replace(k, v)
    return fixture


def _get_target() -> EcoforestApi:
    """Return a mock Envoy."""
    return EcoforestApi("http://127.0.0.1")


@pytest.mark.asyncio
@respx.mock
async def test_get():
    """Get status information."""
    target = _get_target()
    respx.post(path=URL_CGI, data={"idOperacion": API_STATUS_OP}).mock(
        return_value=httpx.Response(200, text=_load_fixture("op-1002-status.txt"))
    )
    respx.post(path=URL_CGI, data={"idOperacion": API_STATS_OP}).mock(
        return_value=httpx.Response(200, text=_load_fixture("op-1020-stats.txt"))
    )
    respx.post(path=URL_CGI, data={"idOperacion": API_ALARMS_OP}).mock(
        return_value=httpx.Response(200, text=_load_fixture("op-1079-alarms.txt"))
    )
    actual = await target.get()
    assert actual is not None
    assert actual == Device(
        is_supported=True,
        model="CC2014_v2",
        model_name="Cordoba glass",
        firmware="30Abr19_v2z",
        serial_number="000025568680000",
        operation_mode=OperationMode.POWER,
        on=False,
        state=State.OFF,
        power=3,
        temperature=20.5,
        alarm=None,
        alarm_code=None,
        environment_temperature=23.5,
        cpu_temperature=32.3,
        gas_temperature=28.1,
        ntc_temperature=25,
        depression=2,
        working_hours=6826,
        ignitions=1152,
        live_pulse=0.0,
        pulse_offset=0.0,
        working_state=0,
        extractor=0,
        working_level=0,
        convecto_air_flow=0.0,
    )


@pytest.mark.asyncio
@respx.mock
@pytest.mark.parametrize(
    ("side_effect", "expected", "message"),
    [
        (
            httpx.Response(401),
            EcoforestAuthenticationRequired,
            "401",
        ),
        (
            httpx.TimeoutException("timeout"),
            EcoforestConnectionError,
            "Timeout occurred while connecting to the device.",
        ),
        (
            httpx.Response(500),
            EcoforestConnectionError,
            "Error occurred while communicating with device.",
        ),
    ],
)
async def test_get_errors(side_effect, expected, message):
    """Get status information with error."""
    target = _get_target()
    respx.post(path=URL_CGI, data={"idOperacion": API_STATUS_OP}).mock(
        side_effect=side_effect
    )
    respx.post(path=URL_CGI, data={"idOperacion": API_STATS_OP}).mock(
        side_effect=side_effect
    )
    respx.post(path=URL_CGI, data={"idOperacion": API_ALARMS_OP}).mock(
        side_effect=side_effect
    )

    with pytest.raises(expected) as err:
        await target.get()
    assert str(err.value) == message


@pytest.mark.asyncio
@respx.mock
async def test_set_temperature():
    """Set target temperature."""
    target = _get_target()
    respx.post(path=URL_CGI, data={"idOperacion": API_STATUS_OP}).mock(
        return_value=httpx.Response(
            200,
            text=_mutate_fixture(
                "op-1002-status.txt",
                [("consigna_temperatura=20.5", "consigna_temperatura=23.5")],
            ),
        )
    )
    respx.post(path=URL_CGI, data={"idOperacion": API_STATS_OP}).mock(
        return_value=httpx.Response(200, text=_load_fixture("op-1020-stats.txt"))
    )
    respx.post(path=URL_CGI, data={"idOperacion": API_ALARMS_OP}).mock(
        return_value=httpx.Response(200, text=_load_fixture("op-1079-alarms.txt"))
    )
    respx.post(
        path=URL_CGI, data={"idOperacion": API_SET_TEMP_OP, "temperatura": 23.5}
    ).mock(return_value=httpx.Response(200, text=_load_fixture("op-1019-set-temp.txt")))

    actual = await target.set_temperature(23.5)
    assert actual is not None
    assert actual == Device(
        is_supported=True,
        model="CC2014_v2",
        model_name="Cordoba glass",
        firmware="30Abr19_v2z",
        serial_number="000025568680000",
        operation_mode=OperationMode.POWER,
        on=False,
        state=State.OFF,
        power=3,
        temperature=23.5,
        alarm=None,
        alarm_code=None,
        environment_temperature=23.5,
        cpu_temperature=32.3,
        gas_temperature=28.1,
        ntc_temperature=25,
        depression=2,
        working_hours=6826,
        ignitions=1152,
        live_pulse=0.0,
        pulse_offset=0.0,
        working_state=0,
        extractor=0,
        working_level=0,
        convecto_air_flow=0.0,
    )


@pytest.mark.asyncio
@respx.mock
async def test_set_power():
    """Set target power."""
    target = _get_target()
    respx.post(path=URL_CGI, data={"idOperacion": API_STATUS_OP}).mock(
        return_value=httpx.Response(
            200,
            text=_mutate_fixture(
                "op-1002-status.txt", [("consigna_potencia=3", "consigna_potencia=5")]
            ),
        )
    )
    respx.post(path=URL_CGI, data={"idOperacion": API_STATS_OP}).mock(
        return_value=httpx.Response(200, text=_load_fixture("op-1020-stats.txt"))
    )
    respx.post(path=URL_CGI, data={"idOperacion": API_ALARMS_OP}).mock(
        return_value=httpx.Response(200, text=_load_fixture("op-1079-alarms.txt"))
    )
    respx.post(
        path=URL_CGI, data={"idOperacion": API_SET_POWER_OP, "potencia": 5}
    ).mock(
        return_value=httpx.Response(200, text=_load_fixture("op-1004-set-power.txt"))
    )

    actual = await target.set_power(5)
    assert actual is not None
    assert actual == Device(
        is_supported=True,
        model="CC2014_v2",
        model_name="Cordoba glass",
        firmware="30Abr19_v2z",
        serial_number="000025568680000",
        operation_mode=OperationMode.POWER,
        on=False,
        state=State.OFF,
        power=5,
        temperature=20.5,
        alarm=None,
        alarm_code=None,
        environment_temperature=23.5,
        cpu_temperature=32.3,
        gas_temperature=28.1,
        ntc_temperature=25,
        depression=2,
        working_hours=6826,
        ignitions=1152,
        live_pulse=0.0,
        pulse_offset=0.0,
        working_state=0,
        extractor=0,
        working_level=0,
        convecto_air_flow=0.0,
    )


@pytest.mark.asyncio
@respx.mock
async def test_turn():
    """Turn on device status."""
    target = _get_target()
    respx.post(path=URL_CGI, data={"idOperacion": API_STATUS_OP}).mock(
        return_value=httpx.Response(
            200,
            text=_mutate_fixture(
                "op-1002-status.txt",
                [("on_off=0", "on_off=1"), ("estado=0", "estado=1")],
            ),
        )
    )
    respx.post(path=URL_CGI, data={"idOperacion": API_STATS_OP}).mock(
        return_value=httpx.Response(200, text=_load_fixture("op-1020-stats.txt"))
    )
    respx.post(path=URL_CGI, data={"idOperacion": API_ALARMS_OP}).mock(
        return_value=httpx.Response(200, text=_load_fixture("op-1079-alarms.txt"))
    )
    respx.post(path=URL_CGI, data={"idOperacion": API_SET_STATE_OP, "on_off": 1}).mock(
        return_value=httpx.Response(200, text=_load_fixture("op-1004-set-power.txt"))
    )

    actual = await target.turn(True)
    assert actual is not None
    assert actual == Device(
        is_supported=True,
        model="CC2014_v2",
        model_name="Cordoba glass",
        firmware="30Abr19_v2z",
        serial_number="000025568680000",
        operation_mode=OperationMode.POWER,
        on=True,
        state=State.STARTING,
        power=3,
        temperature=20.5,
        alarm=None,
        alarm_code=None,
        environment_temperature=23.5,
        cpu_temperature=32.3,
        gas_temperature=28.1,
        ntc_temperature=25,
        depression=2,
        working_hours=6826,
        ignitions=1152,
        live_pulse=0.0,
        pulse_offset=0.0,
        working_state=0,
        extractor=0,
        working_level=0,
        convecto_air_flow=0.0,
    )



import json

import httpx

ENDPOINT_PROBE_EXCEPTIONS = (json.JSONDecodeError, httpx.HTTPError)


class EcoforestError(Exception):
    """Base class for Ecoforest exceptions."""


class EcoforestConnectionError(EcoforestError):
    """Ecoforest connection exception."""


class EcoforestAuthenticationRequired(EcoforestError):
    """Exception raised when authentication failed."""

    def __init__(self, status: str) -> None:
        self.status = status



import logging
from http import HTTPStatus
from typing import Any

import httpx

from pyecoforest.models.device import Device

from .const import (
    API_ALARMS_OP,
    API_SET_POWER_OP,
    API_SET_STATE_OP,
    API_SET_TEMP_OP,
    API_STATS_OP,
    API_STATUS_OP,
    LOCAL_TIMEOUT,
    URL_CGI,
)
from .exceptions import EcoforestAuthenticationRequired, EcoforestConnectionError
from .ssl import NO_VERIFY_SSL_CONTEXT

_LOGGER = logging.getLogger(__name__)


class EcoforestApi:
    """Class for communicating with an ecoforest device."""

    def __init__(
        self,
        host: str,
        auth: httpx.BasicAuth | None = None,
        client: httpx.AsyncClient | None = None,
        timeout: float | httpx.Timeout | None = None,
    ) -> None:
        self._host = host
        self._auth = auth
        # We use our own httpx client session so we can disable SSL verification,
        # the device use self-signed SSL certs.
        self._timeout = timeout or LOCAL_TIMEOUT
        self._client = client or httpx.AsyncClient(
            base_url=self._host, verify=NO_VERIFY_SSL_CONTEXT
        )  # nosec

    async def get(self) -> Device:
        """Retrieve ecoforest information from api."""
        return Device.build(
            {
                "status": await self._status(),
                "stats": await self._stats(),
                "alarms": await self._alarms(),
            }
        )

    async def turn(self, on: bool | None = False) -> Device:
        """Turn device on and off."""
        await self._request(
            data={"idOperacion": API_SET_STATE_OP, "on_off": 1 if on else 0}
        )
        return await self.get()

    async def set_temperature(self, target: float) -> Device:
        """Set device target temperature."""
        await self._request(
            data={"idOperacion": API_SET_TEMP_OP, "temperatura": target}
        )
        return await self.get()

    async def set_power(self, target: int) -> Device:
        """Set device target power."""
        await self._request(data={"idOperacion": API_SET_POWER_OP, "potencia": target})
        return await self.get()

    async def _request(self, data: dict[str, Any] | None = None) -> httpx.Response:
        """Make a request to the device."""
        if _LOGGER.isEnabledFor(logging.DEBUG):
            _LOGGER.debug("Sending POST to %s with data %s", URL_CGI, data)

        try:
            response = await self._client.post(
                URL_CGI,
                auth=self._auth,
                timeout=self._timeout,
                data=data,
            )
            response.raise_for_status()
        except httpx.TimeoutException as error:
            raise EcoforestConnectionError(
                "Timeout occurred while connecting to the device."
            ) from error
        except httpx.HTTPError as error:
            if error.response.status_code in (
                HTTPStatus.UNAUTHORIZED,
                HTTPStatus.FORBIDDEN,
            ):
                raise EcoforestAuthenticationRequired(
                    error.response.status_code
                ) from error
            raise EcoforestConnectionError(
                "Error occurred while communicating with device."
            ) from error

        parsed = self._parse(response.text)

        if _LOGGER.isEnabledFor(logging.DEBUG):
            _LOGGER.debug("Received from POST with data %s", parsed)

        return parsed

    async def _status(self) -> dict[str, str]:
        """Retrieve ecoforest status."""
        return await self._request(data={"idOperacion": API_STATUS_OP})

    async def _stats(self) -> dict[str, str]:
        """Retrieve ecoforest stats."""
        return await self._request(data={"idOperacion": API_STATS_OP})

    async def _alarms(self) -> dict[str, str]:
        """Retrieve ecoforest information from api."""
        return await self._request(
            data={
                "idOperacion": API_ALARMS_OP,
            }
        )

    def _parse(self, response: str) -> dict[str, str]:
        """Parse request data and return as dictionary."""
        # discard last line ?
        reply = {}
        for e in response.split("\n")[:-1]:
            pair = e.split("=")
            # discard lines without pairs
            if len(pair) == 2:
                reply[pair[0]] = pair[1]
        # Remove all white spaces from bad response from ecoforest ...
        reply = {x.translate({32: None}): y for x, y in reply.items()}
        return reply



import pytest

from pyecoforest.exceptions import EcoforestError
from pyecoforest.models.device import Alarm, Device, OperationMode, State


def get_api_data():
    return {
        "status": {
            "on_off": "0",
            "estado": "0",
            "consigna_potencia": "6",
            "consigna_temperatura": "22",
            "temperatura": "24.5",
            "modo_operacion": "0",
        },
        "stats": {
            "Me": "model-version",
            "Vs": "firmware-version",
            "Ns": "serial-number",
            "Tp": "33.5",
            "Th": "36.5",
            "Tn": "23.5",
            "Da": "002",
            "Nh": "000006826",
            "Ne": "001152",
            "Pn": "0.0",
            "Pf": "0.0",
            "Es": "0",
            "Ex": "0.0",
            "Ni": "0",
            "Co": "0.0",
        },
        "alarms": {"get_alarmas": "A099"},
    }


def test_operation_mode_build():
    assert OperationMode.build("0") == OperationMode.POWER
    assert OperationMode.build("1") == OperationMode.TEMPERATURE
    assert OperationMode.build("2") == OperationMode.EMERGENCY
    with pytest.raises(EcoforestError) as error:
        OperationMode.build("3")
    assert str(error.value) == "The operation mode 3 is not a valid operation!"


def test_state_build():
    assert State.build("0") == State.OFF
    for i in [*range(1, 5), 10]:
        assert State.build(i) == State.STARTING
    for i in range(5, 7):
        assert State.build(i) == State.PRE_HEATING
    assert State.build("7") == State.ON
    for i in [8, 11, -3]:
        assert State.build(i) == State.SHUTTING_DOWN
    assert State.build("-20") == State.STAND_BY
    assert State.build("-4") == State.ALARM
    with pytest.raises(EcoforestError) as error:
        State.build("9")
    assert str(error.value) == "The state 9 is not a valid state!"


def test_alarm_build():
    for i in ["A001", "A002"]:
        assert Alarm.build(i) == Alarm.AIR_DEPRESSION
    assert Alarm.build("A012") == Alarm.CPU_OVERHEATING
    assert Alarm.build("A099") == Alarm.PELLETS
    assert Alarm.build("N") is None
    assert Alarm.build("A100") == Alarm.UNKNOWN


def test_device_build():
    assert Device.build(get_api_data()) == Device(
        is_supported=False,
        model="model-version",
        model_name="Cordoba glass",
        firmware="firmware-version",
        serial_number="serial-number",
        operation_mode=OperationMode.POWER,
        on=False,
        state=State.OFF,
        power=6,
        temperature=22,
        alarm=Alarm.PELLETS,
        alarm_code="A099",
        environment_temperature=24.5,
        cpu_temperature=33.5,
        gas_temperature=36.5,
        ntc_temperature=23.5,
        depression=2,
        working_hours=6826,
        ignitions=1152,
        live_pulse=0.0,
        pulse_offset=0.0,
        working_state=0,
        extractor=0,
        working_level=0,
        convecto_air_flow=0.0,
    )


def test_device_build_with_supportted_device():
    data = get_api_data()
    data["stats"]["Me"] = "CC2014_v2"
    assert Device.build(data).is_supported is True
